@@ -1,3 +1,5 @@
+import re
+
 import django
 import pytest
 from crispy_forms import __version__
@@ -24,13 +26,12 @@ from crispy_forms.utils import render_crispy_form
 from django import forms
 from django.template import Context, Template
 from django.test import override_settings
-from django.utils.translation import activate, deactivate
-from django.utils.translation import gettext as _
 
 from .forms import (
     CheckboxesSampleForm,
     CustomCheckboxSelectMultiple,
     CustomRadioSelect,
+    GroupedChoiceForm,
     SampleForm,
     SampleFormCustomWidgets,
 )
@@ -43,17 +44,6 @@ CONVERTERS = {
 }
 
 
-def test_field_with_custom_template():
-    test_form = SampleForm()
-    test_form.helper = FormHelper()
-    test_form.helper.layout = Layout(
-        Field("email", template="custom_field_template.html")
-    )
-
-    html = render_crispy_form(test_form)
-    assert "<h1>Special custom field</h1>" in html
-
-
 def test_multiwidget_field():
     template = Template(
         """
@@ -119,35 +109,6 @@ def test_field_wrapper_class():
     assert html.count('class="form-group testing"') == 1
 
 
-def test_html_with_carriage_returns():
-    test_form = SampleForm()
-    test_form.helper = FormHelper()
-    test_form.helper.layout = Layout(
-        HTML(
-            """
-            if (a==b){
-                // some comment
-                a+1;
-                foo();
-            }
-        """
-        )
-    )
-    html = render_crispy_form(test_form)
-    assert html.count("\n") == 27
-
-
-def test_i18n():
-    activate("es")
-    form = SampleForm()
-    form.helper = FormHelper()
-    form.helper.layout = Layout(HTML(_("Enter a valid value.")))
-    html = render_crispy_form(form)
-    assert "Introduzca un valor válido" in html
-
-    deactivate()
-
-
 def test_remove_labels():
     form = SampleForm()
     # remove boolean field as label is still printed in boostrap
@@ -162,7 +123,7 @@ def test_remove_labels():
 
 
 class TestBootstrapLayoutObjects:
-    def test_custom_django_widget(self):
+    def test_custom_django_widget(self, settings):
         # Make sure an inherited RadioSelect gets rendered as it
         form = SampleFormCustomWidgets()
         assert isinstance(form.fields["inline_radios"].widget, CustomRadioSelect)
@@ -170,7 +131,10 @@ class TestBootstrapLayoutObjects:
         form.helper.layout = Layout("inline_radios")
 
         html = render_crispy_form(form)
-        assert 'class="radio"' in html
+        if settings.CRISPY_TEMPLATE_PACK == "bootstrap4":
+            assert 'class="custom-control-input"' in html
+        else:
+            assert 'class="radio"' in html
 
         # Make sure an inherited CheckboxSelectMultiple gets rendered as it
         assert isinstance(
@@ -178,10 +142,13 @@ class TestBootstrapLayoutObjects:
         )
         form.helper.layout = Layout("checkboxes")
         html = render_crispy_form(form)
-        assert 'class="checkbox"' in html
+        if settings.CRISPY_TEMPLATE_PACK == "bootstrap4":
+            assert 'class="custom-control-input"' in html
+        else:
+            assert 'class="checkbox"' in html
 
     @override_settings(CRISPY_CLASS_CONVERTERS=CONVERTERS)
-    def test_prepended_appended_text(self):
+    def test_prepended_appended_text(self, settings):
         test_form = SampleForm()
         test_form.helper = FormHelper()
         test_form.helper.layout = Layout(
@@ -191,24 +158,67 @@ class TestBootstrapLayoutObjects:
             AppendedText("password1", "#", css_class="input-lg"),
             PrependedText("password2", "$", css_class="input-sm"),
         )
-        if django.VERSION >= (5, 0):
-            # Added 'aria-describedby' for fields with help_text
-            # https://docs.djangoproject.com/en/5.0/releases/5.0/#forms
-            expected = parse_expected(
-                "bootstrap3/test_layout_objects/test_prepended_appended_text_gte50.html"
+        if django.VERSION < (5, 0):
+            expected = (
+                "bootstrap4/test_layout_objects/test_prepended_appended_text_lt50.html"
             )
         else:
-            expected = parse_expected(
-                "bootstrap3/test_layout_objects/test_prepended_appended_text.html"
+            expected = (
+                "bootstrap4/test_layout_objects/test_prepended_appended_text.html"
             )
-        assert parse_form(test_form) == expected
+        assert parse_form(test_form) == parse_expected(expected)
+
+    def test_prepended_wrapper_class(self):
+        test_form = SampleForm()
+        test_form.helper = FormHelper()
+        test_form.helper.layout = Layout(
+            PrependedAppendedText(
+                "email", "@", "gmail.com", wrapper_class="wrapper class"
+            ),
+            PrependedAppendedText("email", "@", "gmail.com"),
+        )
+        html = render_crispy_form(test_form)
+        assert html.count('<div id="div_id_email" class="form-group">') == 1
+        assert (
+            html.count('<div id="div_id_email" class="form-group wrapper class">') == 1
+        )
+
+    def test_prepended_appended_text_in_select(self):
+        test_form = SampleForm()
+        test_form.fields["select"] = forms.ChoiceField(
+            label="Select field", choices=[(1, "Choice 1"), (2, "Choice 2")]
+        )
+
+        test_form.helper = FormHelper()
+        test_form.helper.layout = Layout(
+            AppendedText("select", "USD"),
+        )
+        html = render_crispy_form(test_form)
+
+        assert html.count("custom-select") == 1
+        assert html.count("USD") == 1
+
+    def test_prepended_appended_text_input_size(self):
+        test_form = SampleForm()
+        test_form.helper = FormHelper()
+        test_form.helper.layout = Layout(
+            PrependedAppendedText(
+                "email", "@", "gmail.com", input_size="input-group-lg"
+            ),
+            AppendedText("password1", "#", input_size="input-group-sm"),
+            PrependedText("password2", "$", input_size="input-group-lg"),
+        )
+        html = render_crispy_form(test_form)
+        assert html.count('<div class="input-group input-group-lg">') == 2
+        assert html.count('<div class="input-group input-group-sm">') == 1
 
     def test_inline_radios(self):
         test_form = CheckboxesSampleForm()
         test_form.helper = FormHelper()
         test_form.helper.layout = Layout(InlineRadios("inline_radios"))
         html = render_crispy_form(test_form)
-        assert html.count('radio-inline"') == 2
+
+        assert html.count('custom-control-inline"') == 2
 
     def test_accordion_and_accordiongroup(self):
         test_form = SampleForm()
@@ -221,9 +231,16 @@ class TestBootstrapLayoutObjects:
         )
         html = render_crispy_form(test_form)
 
-        assert html.count('<div class="panel panel-default"') == 2
-        assert html.count('<div class="panel-group"') == 1
-        assert html.count('<div class="panel-heading">') == 2
+        match = re.search('div id="(accordion-\\d+)"', html)
+        assert match
+
+        accordion_id = match.group(1)
+
+        assert html.count('<div class="card mb-2"') == 2
+        assert html.count('<div class="card-header"') == 2
+
+        assert html.count('data-parent="#{}"'.format(accordion_id)) == 2
+
         assert html.count('<div id="one"') == 1
         assert html.count('<div id="two"') == 1
         assert html.count('name="first_name"') == 1
@@ -243,7 +260,7 @@ class TestBootstrapLayoutObjects:
         # The first time, there should be one of them there.
         html = render_crispy_form(test_form)
 
-        accordion_class = "panel-collapse collapse in"
+        accordion_class = "collapse show"
 
         assert html.count('<div id="one" class="%s"' % accordion_class) == 1
 
@@ -294,13 +311,13 @@ class TestBootstrapLayoutObjects:
 
         assert (
             html.count(
-                '<ul class="nav nav-tabs"> <li class="tab-pane active">'
-                '<a href="#custom-name" data-toggle="tab">One</a></li>'
+                '<ul class="nav nav-tabs"> <li class="nav-item">'
+                '<a class="nav-link active" href="#custom-name" data-toggle="tab">'
+                "One</a></li>"
             )
             == 1
         )
-        assert html.count('<li class="tab-pane') == 2
-        assert html.count("tab-pane") == 4
+        assert html.count("tab-pane") == 2
 
         assert html.count('class="tab-pane first-tab-class active"') == 1
 
@@ -357,7 +374,7 @@ class TestBootstrapLayoutObjects:
         assert 'class="second"' in html
 
     @override_settings(CRISPY_CLASS_CONVERTERS=CONVERTERS)
-    def test_field_with_buttons(self):
+    def test_field_with_buttons(self, settings):
         form = SampleForm()
         form.helper = FormHelper()
         form.helper.layout = Layout(
@@ -374,7 +391,7 @@ class TestBootstrapLayoutObjects:
             )
         )
         assert parse_form(form) == parse_expected(
-            "bootstrap3/test_layout_objects/test_field_with_buttons.html"
+            "bootstrap4/test_layout_objects/test_field_with_buttons.html"
         )
 
     def test_hidden_fields(self):
@@ -406,8 +423,8 @@ class TestBootstrapLayoutObjects:
         test_form.helper[1].wrap(InlineCheckboxes, inline=True)
         html = render_crispy_form(test_form)
 
-        assert html.count('inline="True"') == 4
-        assert html.count('checkbox-inline"') == 3
+        assert html.count('inline="True"') == 3
+        assert html.count('custom-control-inline"') == 3
 
     def test_multiple_checkboxes_unique_ids(self):
         test_form = CheckboxesSampleForm()
@@ -428,6 +445,40 @@ class TestBootstrapLayoutObjects:
             expected_str = f'id="id_{id_suffix}"'
             assert html.count(expected_str) == 1
 
+    def test_grouped_checkboxes_radios(self):
+        form = GroupedChoiceForm()
+        form.helper = FormHelper()
+        form.helper.layout = Layout("checkbox_select_multiple")
+        assert parse_form(form) == parse_expected(
+            "bootstrap4/test_layout_objects/test_grouped_checkboxes.html"
+        )
+        form.helper.layout = Layout("radio")
+        assert parse_form(form) == parse_expected(
+            "bootstrap4/test_layout_objects/test_grouped_radios.html"
+        )
+
+        form = GroupedChoiceForm({})
+        form.helper = FormHelper()
+        form.helper.layout = Layout("checkbox_select_multiple")
+        if django.VERSION < (5, 0):
+            expected = (
+                "bootstrap4/test_layout_objects/"
+                "test_grouped_checkboxes_failing_lt50.html"
+            )
+        else:
+            expected = (
+                "bootstrap4/test_layout_objects/test_grouped_checkboxes_failing.html"
+            )
+        assert parse_form(form) == parse_expected(expected)
+        form.helper.layout = Layout("radio")
+        if django.VERSION < (5, 0):
+            expected = (
+                "bootstrap4/test_layout_objects/test_grouped_radios_failing_lt50.html"
+            )
+        else:
+            expected = "bootstrap4/test_layout_objects/test_grouped_radios_failing.html"
+        assert parse_form(form) == parse_expected(expected)
+
     def test_non_ascii_chars_in_container_name(self):
         """
         Test if non-ASCII characters are saved as css_id property.
@@ -436,16 +487,17 @@ class TestBootstrapLayoutObjects:
         test_container = Container(name, "val1", "val2")
         assert test_container.css_id == name
 
-    def test_modal_no_kwargs(self):
+    def test_bs4_modal_no_kwargs(self):
         form = SampleForm()
         form.helper = FormHelper()
         form.helper.layout = Layout(Modal(Field("first_name")))
 
+        print(parse_form(form))
         assert parse_form(form) == parse_expected(
-            "bootstrap3/test_layout_objects/bootstrap_modal_no_kwargs.html"
+            "bootstrap4/test_layout_objects/bootstrap_modal_no_kwargs.html"
         )
 
-    def test_modal_with_kwargs(self):
+    def test_bs4_modal_with_kwargs(self):
         form = SampleForm()
         form.helper = FormHelper()
         form.helper.layout = Layout(
@@ -460,14 +512,14 @@ class TestBootstrapLayoutObjects:
         )
 
         assert parse_form(form) == parse_expected(
-            "bootstrap3/test_layout_objects/bootstrap_modal_with_kwargs.html"
+            "bootstrap4/test_layout_objects/bootstrap_modal_with_kwargs.html"
         )
 
     @pytest.mark.skipif(
         __version__[0] == "1",
         reason="#1230 is post 1.x and fixed double class attributes bug",
     )
-    def test_FormActions(self):
+    def test_FormActions(self, settings):
         form = SampleForm()
         form.helper = FormHelper()
         form.helper.field_class = "field-class"
@@ -481,6 +533,7 @@ class TestBootstrapLayoutObjects:
             ),
         )
 
+        template_pack = settings.CRISPY_TEMPLATE_PACK
         assert parse_form(form) == parse_expected(
-            "bootstrap3/test_layout_objects/test_FormActions.html"
+            f"{template_pack}/test_layout_objects/test_FormActions.html"
         )
